# Copyright 2023 sepes authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import dataclasses as dc
import inspect
import os
from typing import Any

import numpy.testing as npt
import pytest

from sepes._src.backend import backend, treelib
from sepes._src.code_build import (
    autoinit,
    build_field_map,
    convert_hints_to_fields,
    field,
    fields,
)
from sepes._src.tree_base import TreeClass
from sepes._src.tree_mask import tree_mask
from sepes._src.tree_util import is_tree_equal, partial, value_and_tree

freeze = lambda x: tree_mask(x, cond=lambda _: True)

test_arraylib = os.environ.get("SEPES_TEST_ARRAYLIB", "numpy")
if test_arraylib == "jax":
    import jax.numpy as arraylib
elif test_arraylib in ["numpy", "default"]:
    import numpy as arraylib
elif test_arraylib == "torch":
    import torch as arraylib

    arraylib.array = arraylib.tensor
else:
    raise ImportError("no backend installed")


def test_fields():
    @autoinit
    class Test(TreeClass):
        a: int = field(default=1, metadata={"meta": 1})
        b: int = 2

    assert len(fields(Test)) == 2
    assert fields(Test)[0].metadata == {"meta": 1}

    with pytest.raises(ValueError):
        field(kind="WRONG")


def test_field():
    assert field(default=1).default == 1

    with pytest.raises(TypeError):
        field(metadata=1)

    @autoinit
    class Test(TreeClass):
        a: int = field(default=1, kind="POS_ONLY")
        b: int = 2

    with pytest.raises(TypeError):
        # positonal only for a
        Test(a=1, b=2)

    assert Test(1, b=2).a == 1
    assert Test(1, 2).b == 2

    @autoinit
    class Test(TreeClass):
        a: int = field(default=1, kind="POS_ONLY")
        b: int = field(default=2, kind="POS_ONLY")

    assert Test(1, 2).a == 1
    assert Test(1, 2).b == 2

    # keyword only
    @autoinit
    class Test(TreeClass):
        a: int = field(default=1, kind="KW_ONLY")
        b: int = 2

    with pytest.raises(TypeError):
        Test(1, 2)

    with pytest.raises(TypeError):
        Test(1, b=2)

    assert Test(a=1, b=2).a == 1

    @autoinit
    class Test(TreeClass):
        a: int = field(default=1, kind="POS_ONLY")
        b: int = field(default=2, kind="KW_ONLY")

    with pytest.raises(TypeError):
        Test(1, 2)

    assert Test(1, b=2).b == 2

    with pytest.raises(TypeError):
        Test(a=1, b=2)

    # test when init is False
    @autoinit
    class Test(TreeClass):
        a: int = field(default=1, init=False, kind="KW_ONLY")
        b: int = 2

    with pytest.raises(TypeError):
        Test(1, 2)

    assert Test(b=2).a == 1


def test_field_alias():
    @autoinit
    class Tree(TreeClass):
        _name: str = field(alias="name")

    tree = Tree(name="test")
    assert tree._name == "test"

    with pytest.raises(TypeError):
        field(alias=1)


def test_field_nondiff():
    class Test(TreeClass):
        def __init__(
            self,
            a=freeze(arraylib.array([1, 2, 3])),
            b=freeze(arraylib.array([4, 5, 6])),
        ):
            self.a = a
            self.b = b

    test = Test()

    assert treelib.flatten(test)[0] == []

    class Test(TreeClass):
        def __init__(self, a=arraylib.array([1, 2, 3]), b=arraylib.array([4, 5, 6])):
            self.a = freeze(a)
            self.b = b

    test = Test()
    npt.assert_allclose(treelib.flatten(test)[0][0], arraylib.array([4, 5, 6]))


def test_post_init():
    @autoinit
    class Test(TreeClass):
        a: int = 1

        def __post_init__(self):
            self.a = 2

    t = Test()

    assert t.a == 2


def test_subclassing():
    @autoinit
    class L0(TreeClass):
        a: int = 1
        b: int = 3
        c: int = 5

        def inc(self, x):
            return x

        def sub(self, x):
            return x - 10

        def __post_init__(self):
            self.c = 5

    @autoinit
    class L1(L0):
        a: int = 2
        b: int = 4

        def __post_init__(self):
            self.d = 5

        def inc(self, x):
            return x + 10

    l1 = L1()

    assert treelib.flatten(l1)[0] == [2, 4, 5, 5]
    assert l1.inc(10) == 20
    assert l1.sub(10) == 0
    assert l1.d == 5

    @autoinit
    class L1(L0):
        a: int = 2
        b: int = 4

    l1 = L1()

    assert treelib.flatten(l1)[0] == [2, 4, 5]


def test_registering_state():
    class L0(TreeClass):
        def __init__(self):
            self.a = 10
            self.b = 20

    t = L0()
    tt = copy.copy(t)

    assert tt.a == 10
    assert tt.b == 20


def test_copy():
    @autoinit
    class L0(TreeClass):
        a: int = 1
        b: int = 3
        c: int = 5

    t = L0()

    assert copy.copy(t).a == 1
    assert copy.copy(t).b == 3
    assert copy.copy(t).c == 5


def test_delattr():
    @autoinit
    class L0(TreeClass):
        a: int = 1
        b: int = 3
        c: int = 5

    t = L0()

    with pytest.raises(AttributeError):
        del t.a

    @autoinit
    class L2(TreeClass):
        a: int = 1

        def delete(self, name):
            del self.a

    t = L2()

    with pytest.raises(AttributeError):
        t.delete("a")


@pytest.mark.skipif(backend == "default", reason="no array")
def test_is_tree_equal():
    assert is_tree_equal(1, 1)
    assert is_tree_equal(1, 2) is False
    assert is_tree_equal(1, 2.0) is False
    assert is_tree_equal([1, 2], [1, 2])

    @autoinit
    class Test1(TreeClass):
        a: int = 1

    @autoinit
    class Test2(TreeClass):
        a: Any = arraylib.array([1, 2, 3])

    assert is_tree_equal(Test1(), Test2()) is False

    assert is_tree_equal(arraylib.array([1, 2, 3]), arraylib.array([1, 2, 3]))
    assert is_tree_equal(arraylib.array([1, 2, 3]), arraylib.array([1, 3, 3])) is False

    @autoinit
    class Test3(TreeClass):
        a: int = 1
        b: int = 2

    assert is_tree_equal(Test1(), Test3()) is False

    assert is_tree_equal(arraylib.array([1, 2, 3]), 1) is False


def test_mutable_field():
    with pytest.raises(TypeError):

        @autoinit
        class Test(TreeClass):
            a: list = [1, 2, 3]

    with pytest.raises(TypeError):

        @autoinit
        class Test2(TreeClass):
            a: list = field(default=[1, 2, 3])


def test_setattr_delattr():
    with pytest.raises(TypeError):

        @autoinit
        class Test(TreeClass):
            def __setattr__(self, k, v):
                pass

    with pytest.raises(TypeError):

        @autoinit
        class _(TreeClass):
            def __delattr__(self, k):
                pass


def test_on_setattr():
    def instance_validator(types):
        def _instance_validator(x):
            if isinstance(x, types) is False:
                raise AssertionError
            return x

        return _instance_validator

    def range_validator(min, max):
        def _range_validator(x):
            if x < min or x > max:
                raise AssertionError
            return x

        return _range_validator

    @autoinit
    class Test(TreeClass):
        a: int = field(on_setattr=[instance_validator(int)])

    with pytest.raises(AssertionError):
        Test(a="a")

    assert Test(a=1).a == 1

    @autoinit
    class Test(TreeClass):
        a: int = field(on_setattr=[instance_validator((int, float))])

    assert Test(a=1).a == 1
    assert Test(a=1.0).a == 1.0

    with pytest.raises(AssertionError):
        Test(a="a")

    @autoinit
    class Test(TreeClass):
        a: int = field(on_setattr=[range_validator(0, 10)])

    with pytest.raises(AssertionError):
        Test(a=-1)

    assert Test(a=0).a == 0

    with pytest.raises(AssertionError):
        Test(a=11)

    @autoinit
    class Test(TreeClass):
        a: int = field(on_setattr=[range_validator(0, 10), instance_validator(int)])

    with pytest.raises(AssertionError):
        Test(a=-1)

    with pytest.raises(AssertionError):
        Test(a=11)

    with pytest.raises(TypeError):

        @autoinit
        class Test(TreeClass):
            a: int = field(on_setattr=1)

    with pytest.raises(TypeError):

        @autoinit
        class Test(TreeClass):
            a: int = field(on_setattr=[1])

    with pytest.raises(TypeError):

        @autoinit
        class Test(TreeClass):
            a: int = field(on_setattr=[lambda: True])

        Test(a=1)


def test_treeclass_frozen_field():
    @autoinit
    class Test(TreeClass):
        a: int = field(on_setattr=[freeze])

    t = Test(1)

    assert t.a == freeze(1)
    assert treelib.flatten(t)[0] == []


def test_super():
    class Test(TreeClass):
        # a:int
        def __init__(self) -> None:
            super().__init__()

    Test()


def test_optional_attrs():
    class Test(TreeClass):
        def __repr__(self):
            return "a"

        def __or__(self, other):
            return 1

    tree = Test()

    assert tree.__repr__() == "a"
    assert tree | tree == 1


def test_override_repr():
    class Test(TreeClass):
        def __repr__(self):
            return "a"

        def __add__(self, other):
            return 1

    assert Test().__repr__() == "a"
    assert Test() + Test() == 1


def test_optional_param():
    @autoinit
    class Test(TreeClass):
        a: int = field(default=1)

    with pytest.raises(TypeError):
        Test() < Test()


def benchmark_dataclass_class():
    count = 10
    annot = {f"leaf_{i}": int for i in range(count)}
    leaves = {f"leaf_{i}": i for i in range(count)}
    Tree = type("Tree", (), {"__annotations__": annot, **leaves})
    Tree = dc.dataclass(Tree)
    return Tree()


def benchmark_treeclass_instance():
    count = 10
    annot = {f"leaf_{i}": int for i in range(count)}
    leaves = {f"leaf_{i}": i for i in range(count)}
    Tree = autoinit(type("Tree", (TreeClass,), {"__annotations__": annot, **leaves}))
    return Tree()


@pytest.mark.benchmark(group="treeclass")
def test_benchmark_treeclass_instance(benchmark):
    benchmark(benchmark_treeclass_instance)


@pytest.mark.benchmark(group="dataclass")
def test_benchmark_dataclass_class(benchmark):
    benchmark(benchmark_dataclass_class)


def test_self_field_name():
    with pytest.raises(ValueError):

        @autoinit
        class Tree(TreeClass):
            self: int = field()


def test_instance_field_map():
    @autoinit
    class Parameter(TreeClass):
        value: Any

    @autoinit
    class Tree(TreeClass):
        bias: int = 0

        def add_param(self, name, param):
            return setattr(self, name, param)

    tree = Tree()

    _, tree_with_weight = value_and_tree(lambda T: T.add_param("weight", Parameter(3)))(
        tree
    )

    assert tree_with_weight.weight == Parameter(3)
    assert "weight" not in vars(tree)


def test_partial():
    def f(a, b, c):
        return a + b + c

    f_a = partial(f, ..., 2, 3)
    assert f_a(1) == 6

    f_b = partial(f, 1, ..., 3)
    assert f_b(2) == 6

    assert f_b == f_b
    assert hash(f_b) == hash(f_b)


def test_kind():
    @autoinit
    class Tree(TreeClass):
        a: int = field(kind="VAR_POS")
        b: int = field(kind="POS_ONLY")
        c: int = field(kind="VAR_KW")
        d: int

    params = dict(inspect.signature(Tree.__init__).parameters)

    assert params["a"].kind is inspect.Parameter.VAR_POSITIONAL
    assert params["b"].kind is inspect.Parameter.POSITIONAL_ONLY
    assert params["c"].kind is inspect.Parameter.VAR_KEYWORD
    assert params["d"].kind is inspect.Parameter.POSITIONAL_OR_KEYWORD

    @autoinit
    class Tree(TreeClass):
        a: int = field(kind="VAR_POS")
        b: int = field(kind="KW_ONLY")

    params = dict(inspect.signature(Tree.__init__).parameters)
    assert params["a"].kind is inspect.Parameter.VAR_POSITIONAL
    assert params["b"].kind is inspect.Parameter.KEYWORD_ONLY

    with pytest.raises(TypeError):

        @autoinit
        class Tree(TreeClass):
            a: int = field(kind="VAR_POS")
            b: int = field(kind="VAR_POS")

    with pytest.raises(TypeError):

        @autoinit
        class Tree(TreeClass):
            a: int = field(kind="VAR_KW")
            b: int = field(kind="VAR_KW")


def test_init_subclass():
    class Test:
        def __init_subclass__(cls, hello):
            cls.hello = hello

    class Test2(TreeClass, Test, hello=1):
        ...

    assert Test2.hello == 1


def test_nested_mutation():
    @autoinit
    class InnerModule(TreeClass):
        a: int = 1

        def f(self):
            self.a += 1
            return self.a

    @autoinit
    class OuterModule(TreeClass):
        inner: InnerModule = InnerModule()

        def ff(self):
            return self.inner.f()

        def df(self):
            del self.inner.a

    _, v = value_and_tree(lambda F: F.ff())(OuterModule())
    assert v.inner.a == 2

    _, v = value_and_tree(lambda F: F.df())(OuterModule())
    assert "a" not in v.inner.__dict__


def test_autoinit_and_user_defined_init():
    with pytest.raises(TypeError):

        @autoinit
        class Tree(TreeClass):
            b: int

            def __init__(self, a):
                self.a = a

        Tree(a=1)

    assert True


def test_nohints():
    assert convert_hints_to_fields(int) is int


def non_field_builder():
    class T:
        ...

    assert dict(build_field_map(T)) == {}


def test_on_getattr():
    @autoinit
    class Tree(TreeClass):
        a: int = field(on_getattr=[lambda x: x + 1])

    assert Tree(a=1).a == 2

    # with subclassing

    @autoinit
    class Parent:
        a: int = field(on_getattr=[lambda x: x + 1])

    class Child(Parent):
        pass

    child = Child(a=1)

    assert child.a == 2
    assert vars(child)["a"] == 1


def test_unannotated_field():
    class T:
        a = field(default=1)

    assert str(T.a.type) == "NULL"


def test_class_var():
    @autoinit
    class T:
        a: int = field(default=1, kind="CLASS_VAR")
        b: int = 2

    assert T.a == 1



# Copyright 2023 sepes authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import functools as ft
import os
from typing import Any

import pytest

from sepes._src.backend import backend, treelib
from sepes._src.code_build import autoinit
from sepes._src.tree_base import TreeClass
from sepes._src.tree_mask import is_masked, tree_mask, tree_unmask
from sepes._src.tree_util import is_tree_equal, leafwise, tree_hash

test_arraylib = os.environ.get("SEPES_TEST_ARRAYLIB", "numpy")
freeze = ft.partial(tree_mask, cond=lambda _: True)
unfreeze = ft.partial(tree_unmask, cond=lambda _: True)

if test_arraylib == "jax":
    import jax.numpy as arraylib
elif test_arraylib in ["numpy", "default"]:
    import numpy as arraylib
elif test_arraylib == "torch":
    import torch as arraylib

    arraylib.array = arraylib.tensor
else:
    raise ImportError("no backend installed")


def test_freeze_unfreeze():
    @autoinit
    class A(TreeClass):
        a: int
        b: int

    a = A(1, 2)
    b = a.at[...].apply(freeze)
    c = (
        a.at["a"]
        .apply(unfreeze, is_leaf=is_masked)
        .at["b"]
        .apply(unfreeze, is_leaf=is_masked)
    )

    assert treelib.flatten(a)[0] == [1, 2]
    assert treelib.flatten(b)[0] == []
    assert treelib.flatten(c)[0] == [1, 2]
    assert unfreeze(freeze(1.0)) == 1.0

    @autoinit
    class A(TreeClass):
        a: int
        b: int

    @autoinit
    class B(TreeClass):
        c: int = 3
        d: A = A(1, 2)

    @autoinit
    class A(TreeClass):
        a: int
        b: int

    a = A(1, 2)
    b = treelib.map(freeze, a)
    c = (
        a.at["a"]
        .apply(unfreeze, is_leaf=is_masked)
        .at["b"]
        .apply(unfreeze, is_leaf=is_masked)
    )

    assert treelib.flatten(a)[0] == [1, 2]
    assert treelib.flatten(b)[0] == []
    assert treelib.flatten(c)[0] == [1, 2]

    @autoinit
    class L0(TreeClass):
        a: int = 0

    @autoinit
    class L1(TreeClass):
        b: L0 = L0()

    @autoinit
    class L2(TreeClass):
        c: L1 = L1()

    t = treelib.map(freeze, L2())

    assert treelib.flatten(t)[0] == []
    assert treelib.flatten(t.c)[0] == []
    assert treelib.flatten(t.c.b)[0] == []

    class L1(TreeClass):
        def __init__(self):
            self.b = L0()

    class L2(TreeClass):
        def __init__(self):
            self.c = L1()

    t = treelib.map(freeze, L2())
    assert treelib.flatten(t.c)[0] == []
    assert treelib.flatten(t.c.b)[0] == []


def test_freeze_errors():
    class T:
        pass

    @autoinit
    class Test(TreeClass):
        a: Any

    t = Test(T())

    # with pytest.raises(Exception):
    t.at[...].set(0)

    with pytest.raises(TypeError):
        t.at[...].apply(arraylib.sin)

    t.at[...].reduce(arraylib.sin)


def test_freeze_with_ops():
    @autoinit
    class A(TreeClass):
        a: int
        b: int

    @autoinit
    class B(TreeClass):
        c: int = 3
        d: A = A(1, 2)

    @autoinit
    class Test(TreeClass):
        a: int = 1
        b: float = freeze(1.0)
        c: str = freeze("test")

    t = Test()
    assert treelib.flatten(t)[0] == [1]

    with pytest.raises(AttributeError):
        treelib.map(freeze, t).a = 1

    with pytest.raises(AttributeError):
        treelib.map(unfreeze, t).a = 1

    hash(t)

    t = Test()
    treelib.map(unfreeze, t, is_leaf=is_masked)
    treelib.map(freeze, t)

    @autoinit
    class Test(TreeClass):
        a: int

    t = treelib.map(freeze, (Test(100)))

    class Test(TreeClass):
        def __init__(self, x):
            self.x = x

    t = Test(arraylib.array([1, 2, 3]))
    assert is_tree_equal(t.at[...].set(None), Test(x=None))

    class T0:
        a: int = 1

    class T1:
        a: T0 = T0()

    t = T1()


def test_freeze_diagram():
    @autoinit
    class A(TreeClass):
        a: int
        b: int

    @autoinit
    class B(TreeClass):
        c: int = 3
        d: A = A(1, 2)

    a = B()
    a = a.at["d"].set(freeze(a.d))
    a = B()

    a = a.at["d"].set(freeze(a.d))  # = a.d.freeze()


def test_freeze_mask():
    @autoinit
    class Test(TreeClass):
        a: int = 1
        b: int = 2
        c: float = 3.0

    t = Test()

    assert treelib.flatten(treelib.map(freeze, t))[0] == []


def test_freeze_nondiff():
    @autoinit
    class Test(TreeClass):
        a: int = freeze(1)
        b: str = "a"

    t = Test()

    assert treelib.flatten(t)[0] == ["a"]
    assert treelib.flatten(treelib.map(freeze, t))[0] == []
    assert treelib.flatten(
        (treelib.map(freeze, t)).at["b"].apply(unfreeze, is_leaf=is_masked)
    )[0] == ["a"]

    @autoinit
    class T0(TreeClass):
        a: Test = Test()

    t = T0()

    assert treelib.flatten(t)[0] == ["a"]
    assert treelib.flatten(treelib.map(freeze, t))[0] == []

    assert treelib.flatten(t)[0] == ["a"]
    assert treelib.flatten(treelib.map(freeze, t))[0] == []


def test_freeze_nondiff_with_mask():
    @autoinit
    class L0(TreeClass):
        a: int = 1
        b: int = 2
        c: int = 3

    @autoinit
    class L1(TreeClass):
        a: int = 1
        b: int = 2
        c: int = 3
        d: L0 = L0()

    @autoinit
    class L2(TreeClass):
        a: int = 10
        b: int = 20
        c: int = 30
        d: L1 = L1()

    t = L2()
    t = t.at["d"]["d"]["a"].apply(freeze)
    t = t.at["d"]["d"]["b"].apply(freeze)

    assert treelib.flatten(t)[0] == [10, 20, 30, 1, 2, 3, 3]


def test_non_dataclass_input_to_freeze():
    assert treelib.flatten(freeze(1))[0] == []


def test_tree_mask():
    @autoinit
    class L0(TreeClass):
        x: int = 2
        y: int = 3

    @leafwise
    @autoinit
    class L1(TreeClass):
        a: int = 1
        b: L0 = L0()

    tree = L1()

    assert treelib.flatten(tree)[0] == [1, 2, 3]
    assert treelib.flatten(treelib.map(freeze, tree))[0] == []
    assert treelib.flatten(treelib.map(freeze, tree))[0] == []
    assert treelib.flatten(tree.at[...].apply(freeze))[0] == []
    assert treelib.flatten(tree.at[tree > 1].apply(freeze))[0] == [1]
    assert treelib.flatten(tree.at[tree == 1].apply(freeze))[0] == [2, 3]
    assert treelib.flatten(tree.at[tree < 1].apply(freeze))[0] == [1, 2, 3]

    assert treelib.flatten(tree.at["a"].apply(freeze))[0] == [2, 3]
    assert treelib.flatten(tree.at["b"].apply(freeze))[0] == [1]
    assert treelib.flatten(tree.at["b"]["x"].apply(freeze))[0] == [1, 3]
    assert treelib.flatten(tree.at["b"]["y"].apply(freeze))[0] == [1, 2]


def test_tree_unmask():
    @autoinit
    class L0(TreeClass):
        x: int = 2
        y: int = 3

    @leafwise
    @autoinit
    class L1(TreeClass):
        a: int = 1
        b: L0 = L0()

    tree = L1()

    frozen_tree = tree.at[...].apply(freeze)
    assert treelib.flatten(frozen_tree)[0] == []

    mask = tree == tree
    unfrozen_tree = frozen_tree.at[mask].apply(unfreeze, is_leaf=is_masked)
    assert treelib.flatten(unfrozen_tree)[0] == [1, 2, 3]

    mask = tree > 1
    unfrozen_tree = frozen_tree.at[mask].apply(unfreeze, is_leaf=is_masked)
    assert treelib.flatten(unfrozen_tree)[0] == [2, 3]

    unfrozen_tree = frozen_tree.at["a"].apply(unfreeze, is_leaf=is_masked)
    # assert treelib.flatten(unfrozen_tree)[0] == [1]

    # unfrozen_tree = frozen_tree.at["b"].apply(unfreeze, is_leaf=is_masked)
    # assert treelib.flatten(unfrozen_tree)[0] == [2, 3]


def test_tree_mask_unfreeze():
    @autoinit
    class L0(TreeClass):
        x: int = 2
        y: int = 3

    @leafwise
    @autoinit
    class L1(TreeClass):
        a: int = 1
        b: L0 = L0()

    tree = L1()

    mask = tree == tree
    frozen_tree = tree.at[...].apply(freeze)
    unfrozen_tree = frozen_tree.at[mask].apply(unfreeze, is_leaf=is_masked)
    assert treelib.flatten(unfrozen_tree)[0] == [1, 2, 3]

    # frozen_tree = tree.at["a"].apply(freeze)
    # unfrozen_tree = frozen_tree.at["a"].apply(unfreeze, is_leaf=is_masked)
    # assert treelib.flatten(unfrozen_tree)[0] == [1, 2, 3]


def test_wrapper():
    # only apply last wrapper
    assert hash((freeze(1))) == tree_hash(1)

    # lhs = _HashableWrapper(1)
    # # test getter
    # assert lhs.__wrapped__ == 1
    # assert lhs.__wrapped__

    # # comparison with the wrapped object
    # assert lhs != 1
    # # hash of the wrapped object
    # assert hash(lhs) == tree_hash(1)

    # test immutability
    frozen_value = freeze(1)

    with pytest.raises(AttributeError):
        frozen_value.__wrapped__ = 2

    assert freeze(1) == freeze(1)
    assert f"{freeze(1)!r}" == "#1"

    wrapped = freeze(1)

    with pytest.raises(AttributeError):
        delattr(wrapped, "__wrapped__")

    assert wrapped != 1


@pytest.mark.skipif(backend == "default", reason="no array backend installed")
def test_tree_mask_tree_unmask():
    tree = [1, 2, 3.0]
    assert treelib.flatten(tree_mask(tree))[0] == [3.0]
    assert treelib.flatten(tree_unmask(tree_mask(tree)))[0] == [1, 2, 3.0]

    mask_func = lambda x: x < 2
    assert treelib.flatten(tree_mask(tree, mask_func))[0] == [2, 3.0]

    assert freeze(freeze(1)) == freeze(1)

    with pytest.raises(TypeError):
        tree_mask({"a": 1}, cond=1.0)

    assert copy.copy(freeze(1)) == freeze(1)

    with pytest.raises(NotImplementedError):
        freeze(1) + 1


@pytest.mark.skipif(backend == "default", reason="no array backend installed")
def test_array_tree_mask_tree_unmask():
    frozen_array = tree_mask(arraylib.ones((5, 5)), cond=lambda _: True)

    assert frozen_array == frozen_array
    assert not (frozen_array == freeze(arraylib.ones((5, 6))))
    # assert not (frozen_array == freeze(arraylib.ones((5, 5)).astype(arraylib.uint8)))
    assert hash(frozen_array) == hash(frozen_array)



# Copyright 2023 sepes authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utility functions for pytrees."""

from __future__ import annotations

import copy
import functools as ft
import operator as op
from math import ceil, floor, trunc
from typing import Any, Callable, Generic, Hashable, Iterator, Sequence, Tuple, TypeVar

from typing_extensions import ParamSpec

import sepes
import sepes._src.backend.arraylib as arraylib
from sepes._src.backend import is_package_avaiable

T = TypeVar("T")
T1 = TypeVar("T1")
T2 = TypeVar("T2")
P = ParamSpec("P")
PyTree = Any
EllipsisType = TypeVar("EllipsisType")
KeyEntry = TypeVar("KeyEntry", bound=Hashable)
TypeEntry = TypeVar("TypeEntry", bound=type)
TraceEntry = Tuple[KeyEntry, TypeEntry]
KeyPath = Tuple[KeyEntry, ...]
TypePath = Tuple[TypeEntry, ...]
KeyTypePath = Tuple[KeyPath, TypePath]


def tree_hash(*trees: PyTree) -> int:
    treelib = sepes._src.backend.treelib
    leaves, treedef = treelib.flatten(trees)
    return hash((*leaves, treedef))


def tree_copy(tree: T) -> T:
    """Return a copy of the tree."""
    # the dispatcher calls copy on the leaves of the tree
    # by default as an extra measure - beside flatten/unflatten-
    # to ensure that the tree is copied completely
    treelib = sepes._src.backend.treelib
    types = tuple(set(tree_copy.copy_dispatcher.registry) - {object})

    def is_leaf(node) -> bool:
        return isinstance(node, types)

    return treelib.map(tree_copy.copy_dispatcher, tree, is_leaf=is_leaf)


# default behavior is to copy the tree elements except for registered types
# like jax arrays which are immutable by default and should not be copied
tree_copy.copy_dispatcher = ft.singledispatch(copy.copy)
tree_copy.def_type = tree_copy.copy_dispatcher.register


@tree_copy.def_type(int)
@tree_copy.def_type(float)
@tree_copy.def_type(complex)
@tree_copy.def_type(str)
@tree_copy.def_type(bytes)
def _(x: T) -> T:
    # skip applying `copy.copy` on immutable atom types
    return x


def is_array_like(node) -> bool:
    return hasattr(node, "shape") and hasattr(node, "dtype")


def _is_leaf_rhs_equal(leaf, rhs):
    if is_array_like(leaf):
        if is_array_like(rhs):
            if leaf.shape != rhs.shape:
                return False
            if leaf.dtype != rhs.dtype:
                return False
            try:
                verdict = arraylib.all(leaf == rhs)
            except NotImplementedError:
                verdict = leaf == rhs
            try:
                return bool(verdict)
            except Exception:
                return verdict  # fail under `jit`
        return False
    return leaf == rhs


def is_tree_equal(*trees: Any) -> bool:
    """Return ``True`` if all pytrees are equal.

    Note:
        trees are compared using their leaves and treedefs.
    """
    treelib = sepes._src.backend.treelib
    tree0, *rest = trees
    leaves0, treedef0 = treelib.flatten(tree0)
    verdict = True

    for tree in rest:
        leaves, treedef = treelib.flatten(tree)
        if (treedef != treedef0) or verdict is False:
            return False
        verdict = ft.reduce(op.and_, map(_is_leaf_rhs_equal, leaves0, leaves), verdict)
    return verdict


class Static(Generic[T]):
    def __init_subclass__(klass, **k) -> None:
        # register subclasses as an empty pytree node
        # written like this to enforce selection of the proper backend
        # every time a subclass is created
        super().__init_subclass__(**k)
        # register with the proper backend
        treelib = sepes._src.backend.treelib
        treelib.register_static(klass)


class partial(ft.partial):
    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        iargs = iter(args)
        args = (next(iargs) if arg is ... else arg for arg in self.args)  # type: ignore
        return self.func(*args, *iargs, **{**self.keywords, **kwargs})


def bcmap(
    func: Callable[P, T],
    broadcast_to: int | str | None = None,
    *,
    is_leaf: Callable[[Any], bool] | None = None,
) -> Callable[P, T]:
    """Map a function over pytree leaves with automatic broadcasting for scalar arguments.

    Args:
        func: the function to be mapped over the pytree.
        broadcast_to: Accepts integer for broadcasting to a specific argument
            or string for broadcasting to a specific keyword argument.
            If ``None``, then the function is broadcasted to the first argument
            or the first keyword argument if no positional arguments are provided.
            Defaults to ``None``.
        is_leaf: a predicate function that returns True if the node is a leaf.

    Example:
        Transform `numpy` functions to work with pytrees:

        >>> import sepes as sp
        >>> import jax.numpy as jnp
        >>> tree_of_arrays = {"a": jnp.array([1, 2, 3]), "b": jnp.array([4, 5, 6])}
        >>> tree_add = sp.bcmap(jnp.add)
        >>> # both lhs and rhs are pytrees
        >>> print(sp.tree_str(tree_add(tree_of_arrays, tree_of_arrays)))
        dict(a=[2 4 6], b=[ 8 10 12])
        >>> # rhs is a scalar
        >>> print(sp.tree_str(tree_add(tree_of_arrays, 1)))
        dict(a=[2 3 4], b=[5 6 7])
    """
    treelib = sepes._src.backend.treelib

    @ft.wraps(func)
    def wrapper(*args, **kwargs):
        cargs = []
        ckwargs = {}
        leaves = []
        kwargs_keys: list[str] = []

        bdcst_to = (
            (0 if len(args) else next(iter(kwargs)))
            if broadcast_to is None
            else broadcast_to
        )

        treedef0 = (
            # reference treedef is the first positional argument
            treelib.flatten(args[bdcst_to], is_leaf=is_leaf)[1]
            if len(args)
            # reference treedef is the first keyword argument
            else treelib.flatten(kwargs[bdcst_to], is_leaf=is_leaf)[1]
        )

        for arg in args:
            if treedef0 == treelib.flatten(arg, is_leaf=is_leaf)[1]:
                cargs += [...]
                leaves += [treedef0.flatten_up_to(arg)]
            else:
                cargs += [arg]

        for key in kwargs:
            if treedef0 == treelib.flatten(kwargs[key], is_leaf=is_leaf)[1]:
                ckwargs[key] = ...
                leaves += [treedef0.flatten_up_to(kwargs[key])]
                kwargs_keys += [key]
            else:
                ckwargs[key] = kwargs[key]

        split_index = len(leaves) - len(kwargs_keys)
        all_leaves = []
        bfunc = partial(func, *cargs, **ckwargs)
        for args_kwargs_values in zip(*leaves):
            args = args_kwargs_values[:split_index]
            kwargs = dict(zip(kwargs_keys, args_kwargs_values[split_index:]))
            all_leaves += [bfunc(*args, **kwargs)]
        return treelib.unflatten(treedef0, all_leaves)

    return wrapper


def swop(func):
    # swaping the arguments of a two-arg function
    return ft.wraps(func)(lambda leaf, rhs: func(rhs, leaf))


def leafwise(klass: type[T]) -> type[T]:
    """A class decorator that adds leafwise operators to a class.

    Leafwise operators are operators that are applied to the leaves of a pytree.
    For example leafwise ``__add__`` is equivalent to:

    - ``tree_map(lambda x: x + rhs, tree)`` if ``rhs`` is a scalar.
    - ``tree_map(lambda x, y: x + y, tree, rhs)`` if ``rhs`` is a pytree
      with the same structure as ``tree``.

    Args:
        klass: The class to be decorated.

    Returns:
        The decorated class.

    Example:
        Use ``numpy`` functions on :class:`TreeClass`` classes decorated with :func:`leafwise`

        >>> import sepes as sp
        >>> import jax.numpy as jnp
        >>> @sp.leafwise
        ... @sp.autoinit
        ... class Point(sp.TreeClass):
        ...    x: float = 0.5
        ...    y: float = 1.0
        ...    description: str = "point coordinates"
        >>> # use :func:`tree_mask` to mask the non-inexact part of the tree
        >>> # i.e. mask the string leaf ``description`` to ``Point`` work
        >>> # with ``jax.numpy`` functions
        >>> co = sp.tree_mask(Point())
        >>> print(sp.bcmap(jnp.where)(co > 0.5, co, 1000))
        Point(x=1000.0, y=1.0, description=#point coordinates)

    Note:
        If a mathematically equivalent operator is already defined on the class,
        then it is not overridden.

    ==================      ============
    Method                  Operator
    ==================      ============
    ``__add__``              ``+``
    ``__and__``              ``&``
    ``__ceil__``             ``math.ceil``
    ``__divmod__``           ``divmod``
    ``__eq__``               ``==``
    ``__floor__``            ``math.floor``
    ``__floordiv__``         ``//``
    ``__ge__``               ``>=``
    ``__gt__``               ``>``
    ``__invert__``           ``~``
    ``__le__``               ``<=``
    ``__lshift__``           ``<<``
    ``__lt__``               ``<``
    ``__matmul__``           ``@``
    ``__mod__``              ``%``
    ``__mul__``              ``*``
    ``__ne__``               ``!=``
    ``__neg__``              ``-``
    ``__or__``               ``|``
    ``__pos__``              ``+``
    ``__pow__``              ``**``
    ``__round__``            ``round``
    ``__sub__``              ``-``
    ``__truediv__``          ``/``
    ``__trunc__``            ``math.trunc``
    ``__xor__``              ``^``
    ==================      ============
    """
    treelib = sepes._src.backend.treelib

    def uop(func):
        def wrapper(self):
            return treelib.map(func, self)

        return ft.wraps(func)(wrapper)

    def bop(func):
        def wrapper(leaf, rhs=None):
            if isinstance(rhs, type(leaf)):
                return treelib.map(func, leaf, rhs)
            return treelib.map(lambda x: func(x, rhs), leaf)

        return ft.wraps(func)(wrapper)

    for key, method in (
        ("__abs__", uop(abs)),
        ("__add__", bop(op.add)),
        ("__and__", bop(op.and_)),
        ("__ceil__", uop(ceil)),
        ("__divmod__", bop(divmod)),
        ("__eq__", bop(op.eq)),
        ("__floor__", uop(floor)),
        ("__floordiv__", bop(op.floordiv)),
        ("__ge__", bop(op.ge)),
        ("__gt__", bop(op.gt)),
        ("__invert__", uop(op.invert)),
        ("__le__", bop(op.le)),
        ("__lshift__", bop(op.lshift)),
        ("__lt__", bop(op.lt)),
        ("__matmul__", bop(op.matmul)),
        ("__mod__", bop(op.mod)),
        ("__mul__", bop(op.mul)),
        ("__ne__", bop(op.ne)),
        ("__neg__", uop(op.neg)),
        ("__or__", bop(op.or_)),
        ("__pos__", uop(op.pos)),
        ("__pow__", bop(op.pow)),
        ("__radd__", bop(swop(op.add))),
        ("__rand__", bop(swop(op.and_))),
        ("__rdivmod__", bop(swop(divmod))),
        ("__rfloordiv__", bop(swop(op.floordiv))),
        ("__rlshift__", bop(swop(op.lshift))),
        ("__rmatmul__", bop(swop(op.matmul))),
        ("__rmod__", bop(swop(op.mod))),
        ("__rmul__", bop(swop(op.mul))),
        ("__ror__", bop(swop(op.or_))),
        ("__round__", bop(round)),
        ("__rpow__", bop(swop(op.pow))),
        ("__rrshift__", bop(swop(op.rshift))),
        ("__rshift__", bop(op.rshift)),
        ("__rsub__", bop(swop(op.sub))),
        ("__rtruediv__", bop(swop(op.truediv))),
        ("__rxor__", bop(swop(op.xor))),
        ("__sub__", bop(op.sub)),
        ("__truediv__", bop(op.truediv)),
        ("__trunc__", uop(trunc)),
        ("__xor__", bop(op.xor)),
    ):
        if key not in vars(klass):
            # do not override any user defined methods
            # this behavior similar is to `dataclasses.dataclass`
            setattr(klass, key, method)
    return klass


def tree_type_path_leaves(
    tree: PyTree,
    *,
    is_leaf: Callable[[Any], bool] | None = None,
    is_path_leaf: Callable[[KeyTypePath], bool] | None = None,
) -> Sequence[tuple[KeyTypePath, Any]]:
    treelib = sepes._src.backend.treelib
    _, atomicdef = treelib.flatten(1)

    # mainly used for visualization
    def flatten_one_level(type_path: KeyTypePath, tree: PyTree):
        # predicate and type path
        if (is_leaf and is_leaf(tree)) or (is_path_leaf and is_path_leaf(type_path)):
            yield type_path, tree
            return

        def one_level_is_leaf(node) -> bool:
            if is_leaf and is_leaf(node):
                return True
            if id(node) == id(tree):
                return False
            return True

        path_leaf, treedef = treelib.path_flatten(tree, is_leaf=one_level_is_leaf)

        if treedef == atomicdef:
            yield type_path, tree
            return

        for key, value in path_leaf:
            keys, types = type_path
            path = ((*keys, *key), (*types, type(value)))
            yield from flatten_one_level(path, value)

    return list(flatten_one_level(((), ()), tree))


class Node:
    # mainly used for visualization
    __slots__ = ["data", "parent", "children", "__weakref__"]

    def __init__(
        self,
        data: tuple[TraceEntry, Any],
        parent: Node | None = None,
    ):
        self.data = data
        self.parent = parent
        self.children: dict[TraceEntry, Node] = {}

    def add_child(self, child: Node) -> None:
        # add child node to this node and set
        # this node as the parent of the child
        if not isinstance(child, Node):
            raise TypeError(f"`child` must be a `Node`, got {type(child)}")
        ti, _ = child.data
        if ti not in self.children:
            # establish parent-child relationship
            child.parent = self
            self.children[ti] = child

    def __iter__(self) -> Iterator[Node]:
        # iterate over children nodes
        return iter(self.children.values())

    def __repr__(self) -> str:
        return f"Node(data={self.data})"

    def __contains__(self, key: TraceEntry) -> bool:
        return key in self.children


def is_path_leaf_depth_factory(depth: int | float):
    # generate `is_path_leaf` function to stop tracing at a certain `depth`
    # in essence, depth is the length of the trace entry
    def is_path_leaf(trace) -> bool:
        keys, _ = trace
        # stop tracing if depth is reached
        return False if depth is None else (depth <= len(keys))

    return is_path_leaf


def construct_tree(
    tree: PyTree,
    is_leaf: Callable[[Any], bool] | None = None,
    is_path_leaf: Callable[[KeyTypePath], bool] | None = None,
) -> Node:
    # construct a tree with `Node` objects using `tree_type_path_leaves`
    # to establish parent-child relationship between nodes

    traces_leaves = tree_type_path_leaves(
        tree,
        is_leaf=is_leaf,
        is_path_leaf=is_path_leaf,
    )

    ti = (None, type(tree))
    vi = tree
    root = Node(data=(ti, vi))

    for trace, leaf in traces_leaves:
        keys, types = trace
        cur = root
        for i, ti in enumerate(zip(keys, types)):
            if ti in cur:
                # common parent node
                cur = cur.children[ti]
            else:
                # new path
                vi = leaf if i == len(keys) - 1 else None
                child = Node(data=(ti, vi))
                cur.add_child(child)
                cur = child
    return root


def value_and_tree(func: Callable[..., T], argnums: int | Sequence[int] = 0):
    """Call a function on copied input argument and return the value and the tree.

    Input arguments are copied before calling the function, and the argument
    specified by ``argnums`` are returned as a tree.

    Args:
        func: A function.
        argnums: The argument number of the tree that will be returned. If multiple
            arguments are specified, the tree will be returned as a tuple.

    Returns:
        A function that returns the value and the tree.

    Example:
        Usage with mutable types:

        >>> import sepes as sp
        >>> mutable_tree = [1, 2, 3]
        >>> def mutating_func(tree):
        ...     tree[0] += 100
        ...     return tree
        >>> new_tree = mutating_func(mutable_tree)
        >>> assert new_tree is mutable_tree
        >>> # now with `value_and_tree` the function does not mutate the tree
        >>> new_tree, _ = sp.value_and_tree(mutating_func)(mutable_tree)
        >>> assert new_tree is not mutable_tree

    Example:
        Usage with immutable types (:class:`.TreeClass`) with support for in-place
        mutation via custom behavior registration using :func:`.value_and_tree.def_mutator`
        and :func:`.value_and_tree.def_immutator`:

        >>> import sepes as sp
        >>> class Counter(sp.TreeClass):
        ...     def __init__(self, count: int):
        ...         self.count = count
        ...     def increment(self, value):
        ...         self.count += value
        ...         return self.count
        >>> counter = Counter(0)
        >>> counter.increment(1)  # doctest: +SKIP
        AttributeError: Cannot set attribute value=1 to `key='count'`  on an immutable instance of `Counter`.
        >>> sp.value_and_tree(lambda counter: counter.increment(1))(counter)
        (1, Counter(count=1))

    Note:
        Use this function on function that:

        - Mutates the input arguments of mutable types (e.g. lists, dicts, etc.).
        - Mutates the input arguments of immutable types that do not support in-place
          mutation and needs special handling that can be registered (e.g. :class:`.TreeClass`)
          using :func:`.value_and_tree.def_mutator` and :func:`.value_and_tree.def_immutator`.

    Note:
        The default behavior of :func:`value_and_tree` is to copy the input
        arguments and then call the function on the copy. However if the function
        mutates some of the input arguments that does not support in-place mutation,
        then the function will fail. In this case, :func:`value_and_tree` enables
        registering custom behavior that modifies the copied input argument to
        allow in-place mutation. and custom function that restores the copied
        argument to its original state after the method call. The following example
        shows how to register custom functions for a simple class that allows
        in-place mutation if ``immutable`` Flag is set to ``False``.

        >>> import jax
        >>> from jax.util import unzip2
        >>> import sepes as sp
        >>> @jax.tree_util.register_pytree_node_class
        ... class MyNode:
        ...     def __init__(self):
        ...         self.counter = 0
        ...         self.immutable = True
        ...     def tree_flatten(self):
        ...         keys, values = unzip2(vars(self).items())
        ...         return tuple(values), tuple(keys)
        ...     @classmethod
        ...     def tree_unflatten(cls, keys, values):
        ...         self = object.__new__(cls)
        ...         vars(self).update(dict(zip(keys, values)))
        ...         return self
        ...     def __setattr__(self, name, value):
        ...         if getattr(self, "immutable", False) is True:
        ...             raise AttributeError("MyNode is immutable")
        ...         object.__setattr__(self, name, value)
        ...     def __repr__(self):
        ...         params = ", ".join(f"{k}={v}" for k, v in vars(self).items())
        ...         return f"MyNode({params})"
        ...     def increment(self) -> None:
        ...         self.counter += 1
        >>> @sp.value_and_tree.def_mutator(MyNode)
        ... def mutable(node) -> None:
        ...     vars(node)["immutable"] = False
        >>> @sp.value_and_tree.def_immutator(MyNode)
        ... def immutable(node) -> None:
        ...     vars(node)["immutable"] = True
        >>> node = MyNode()
        >>> sp.value_and_tree(lambda node: node.increment())(node)
        (None, MyNode(counter=1, immutable=True))
    """
    treelib = sepes._src.backend.treelib
    is_int_argnum = isinstance(argnums, int)
    argnums = [argnums] if is_int_argnum else argnums

    def mutate_is_leaf(node):
        value_and_tree.mutator_dispatcher(node)
        return False

    def immutate_is_leaf(node):
        value_and_tree.immutator_dispatcher(node)
        return False

    @ft.wraps(func)
    def stateless_func(*args, **kwargs) -> tuple[T, PyTree | tuple[PyTree, ...]]:
        # copy the incoming inputs
        (args, kwargs) = tree_copy((args, kwargs))
        # and edit the node/record to make it mutable (if there is a rule for it)
        treelib.map(lambda _: _, (args, kwargs), is_leaf=mutate_is_leaf)
        output = func(*args, **kwargs)
        # traverse each node in the tree depth-first manner
        # to undo the mutation (if there is a rule for it)
        treelib.map(lambda _: _, (args, kwargs), is_leaf=immutate_is_leaf)
        out_args = tuple(a for i, a in enumerate(args) if i in argnums)
        out_args = out_args[0] if is_int_argnum else out_args
        return output, out_args

    return stateless_func


value_and_tree.mutator_dispatcher = ft.singledispatch(lambda node: node)
value_and_tree.immutator_dispatcher = ft.singledispatch(lambda node: node)
value_and_tree.def_mutator = value_and_tree.mutator_dispatcher.register
value_and_tree.def_immutator = value_and_tree.immutator_dispatcher.register


if is_package_avaiable("jax"):
    import jax

    # basically avoid calling copy on jax arrays because they
    # are immutable by default
    @tree_copy.def_type(jax.Array)
    def _(node: jax.Array) -> jax.Array:
        return node

    # avoid calling __copy__ on jitted functions becasue they loses their
    # wrapped function attributes (maybe a bug in jax)
    @tree_copy.def_type(type(jax.jit(lambda x: x)))
    def _(node: T1) -> T1:
        return node



# Copyright 2023 sepes authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import numpy as np
import torch
from torch import Tensor

import sepes._src.backend.arraylib as arraylib

floatings = [torch.float16, torch.float32, torch.float64]
complexes = [torch.complex32, torch.complex64, torch.complex128]
integers = [torch.int8, torch.int16, torch.int32, torch.int64]

arraylib.tobytes.register(Tensor, lambda x: np.from_dlpack(x).tobytes())
arraylib.where.register(Tensor, torch.where)
arraylib.nbytes.register(Tensor, lambda x: x.nbytes)
arraylib.shape.register(Tensor, lambda x: tuple(x.shape))
arraylib.dtype.register(Tensor, lambda x: x.dtype)
arraylib.min.register(Tensor, torch.min)
arraylib.max.register(Tensor, torch.max)
arraylib.mean.register(Tensor, torch.mean)
arraylib.std.register(Tensor, torch.std)
arraylib.all.register(Tensor, torch.all)
arraylib.array_equal.register(Tensor, torch.equal)
arraylib.is_floating.register(Tensor, lambda x: x.dtype in floatings)
arraylib.is_integer.register(Tensor, lambda x: x.dtype in integers)
arraylib.is_inexact.register(Tensor, lambda x: x.dtype in floatings + complexes)
arraylib.is_bool.register(Tensor, lambda x: x.dtype == torch.bool)
arraylib.ndarrays.append(Tensor)



# Copyright 2023 sepes authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import math
import os
from typing import Any

import pytest

from sepes._src.backend import backend
from sepes._src.code_build import autoinit, field
from sepes._src.tree_base import TreeClass
from sepes._src.tree_mask import tree_mask
from sepes._src.tree_util import bcmap, is_tree_equal, leafwise

freeze = lambda x: tree_mask(x, cond=lambda _: True)

test_arraylib = os.environ.get("SEPES_TEST_ARRAYLIB", "numpy")
if test_arraylib == "jax":
    import jax.numpy as arraylib
elif test_arraylib in ["numpy", "default"]:
    import numpy as arraylib
elif test_arraylib == "torch":
    import torch as arraylib

    arraylib.array = arraylib.tensor
else:
    raise ImportError("no backend installed")


@leafwise
@autoinit
class Tree(TreeClass):
    """A simple tree class for math testing."""

    a: float
    b: float
    c: float
    name: str = field(on_setattr=[freeze])  # exclude it from math operations


tree1 = Tree(-10, 20, 30, "A")
tree2 = Tree(arraylib.array([-10]), arraylib.array([20]), arraylib.array([30]), "A")


@pytest.mark.parametrize(
    ["tree", "expected"],
    [
        [abs(tree1), Tree(10, 20, 30, "A")],
        [tree1 + tree1, Tree(-20, 40, 60, "A")],
        [tree1 + 1, Tree(-9, 21, 31, "A")],
        [math.ceil(tree1), Tree(-10, 20, 30, "A")],
        [
            divmod(tree1, tree1),
            Tree(divmod(-10, -10), divmod(20, 20), divmod(30, 30), "A"),
        ],
        [tree1, tree1],
        [math.floor(tree1), Tree(-10, 20, 30, "A")],
        [tree1 // tree1, Tree(-10 // (-10), 20 // (20), 30 // (30), "A")],
        [tree1 // 2, Tree(-5, 10, 15, "A")],
        [tree1 >= tree1, Tree(-10 >= -10, 20 >= 20, 30 >= 30, "A")],
        [tree1 >= 1, Tree(-10 >= 1, 20 >= 1, 30 >= 1, "A")],
        [tree1 > tree1, Tree(-10 > -10, 20 > 20, 30 > 30, "A")],
        [tree1 > 1, Tree(-10 > 1, 20 > 1, 30 > 1, "A")],
        [~tree1, Tree(~-10, ~20, ~30, "A")],
        [tree1 <= tree1, Tree(-10 <= -10, 20 <= 20, 30 <= 30, "A")],
        [tree1 <= 1, Tree(-10 <= 1, 20 <= 1, 30 <= 1, "A")],
        [tree1 << 1, Tree(-10 << 1, 20 << 1, 30 << 1, "A")],
        [tree1 < tree1, Tree(-10 < -10, 20 < 20, 30 < 30, "A")],
        [tree1 != tree1, Tree(-10 != -10, 20 != 20, 30 != 30, "A")],
        [tree1 != 1, Tree(-10 != 1, 20 != 1, 30 != 1, "A")],
        [tree1 or tree1, Tree(-10 or -10, 20 or 20, 30 or 30, "A")],
        [tree1 or 1, Tree(-10 or 1, 20 or 1, 30 or 1, "A")],
        [+tree1, Tree(+(-10), +20, +30, "A")],
        [tree1**2, Tree(100, 400, 900, "A")],
        [1 + tree1, Tree(-9, 21, 31, "A")],
        [1 and tree1, Tree(1 and -10, 1 and 20, 1 and 30, "A")],
        [divmod(1, tree1), Tree(divmod(1, -10), divmod(1, 20), divmod(1, 30), "A")],
        [1 | tree1, Tree(1 | -10, 1 | 20, 1 | 30, "A")],
        [2**tree1, Tree(2**-10, 2**20, 2**30, "A")],
        [1 ^ tree1, Tree(1 ^ -10, 1 ^ 20, 1 ^ 30, "A")],
        [1 - tree1, Tree(1 + 10, 1 - 20, 1 - 30, "A")],
        [1 / tree1, Tree(1 / -10, 1 / 20, 1 / 30, "A")],
        [1 * tree1, Tree(1 * -10, 1 * 20, 1 * 30, "A")],
        [math.trunc(tree1), Tree(-10, 20, 30, "A")],
        [round(tree1), Tree(-10, 20, 30, "A")],
        [tree1 ^ tree1, Tree(-10 ^ -10, 20 ^ 20, 30 ^ 30, "A")],
        [
            tree2 @ tree2,
            Tree(
                arraylib.array([10]) @ arraylib.array([10]),
                arraylib.array([20]) @ arraylib.array([20]),
                arraylib.array([30]) @ arraylib.array([30]),
                "A",
            ),
        ],
    ],
)
def test_tree_math(tree, expected):
    assert is_tree_equal(tree, expected)
    assert tree == expected  # same as above


@leafwise
@autoinit
class Tree(TreeClass):
    """A simple tree class for math testing."""

    a: Any
    b: Any
    c: Any
    d: Any


tree = Tree(a=(10, 20, 30), b=(40, 50, 60), c=(70, 80, 90), d=(100, 110, 120))


def where(condition, x, y):
    # torch expects tensors
    condition = arraylib.array(condition)
    x = arraylib.array(x)
    y = arraylib.array(y)
    return arraylib.where(condition, x, y)


@pytest.mark.skipif(backend == "torch", reason="where needs tensor arguments")
@pytest.mark.parametrize(
    ["tree", "expected"],
    [
        [
            bcmap(where)(tree > 10, 0, tree),
            bcmap(arraylib.array)(
                Tree(a=(10, 0, 0), b=(0, 0, 0), c=(0, 0, 0), d=(0, 0, 0))
            ),
        ],
        [
            bcmap(where)(tree > 10, 0, y=tree),
            bcmap(arraylib.array)(
                Tree(a=(10, 0, 0), b=(0, 0, 0), c=(0, 0, 0), d=(0, 0, 0))
            ),
        ],
        [
            bcmap(where)(tree > 10, x=0, y=tree),
            bcmap(arraylib.array)(
                Tree(a=(10, 0, 0), b=(0, 0, 0), c=(0, 0, 0), d=(0, 0, 0))
            ),
        ],
        [
            bcmap(where)(condition=tree > 10, x=0, y=tree),
            bcmap(arraylib.array)(
                Tree(a=(10, 0, 0), b=(0, 0, 0), c=(0, 0, 0), d=(0, 0, 0))
            ),
        ],
    ],
)
def test_bcmap(tree, expected):
    assert is_tree_equal(tree, expected)


def test_math_operations_errors():
    with pytest.raises(TypeError):
        tree1 + "s"


def test_bcmap_int_argnum_broadcast_to():
    def func(x, y):
        return x + y

    assert bcmap(func, broadcast_to=1)(1, [2, 3, 4]) == [3, 4, 5]


def test_bcmap_key_argnum_broadcast_to():
    def func(x, y):
        return x + y

    assert bcmap(func, broadcast_to="y")(x=1, y=[2, 3, 4]) == [3, 4, 5]



# Copyright 2023 sepes authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import dataclasses as dc
from typing import Any, Callable, Hashable, Iterable

import optree as ot

from sepes._src.backend.treelib import (
    AbstractTreeLib,
    KeyPathLeaf,
    ParallelConfig,
    Tree,
    concurrent_map,
    namespace,
)


@dc.dataclass(frozen=True)
class SequenceKey:
    idx: int

    def __str__(self):
        return f"[{repr(self.idx)}]"


@dc.dataclass(frozen=True)
class DictKey:
    key: Hashable

    def __str__(self):
        return f"[{repr(self.key)}]"


@dc.dataclass(frozen=True)
class GetAttrKey:
    name: str

    def __str__(self):
        return f".{self.name}"


@dc.dataclass(frozen=True)
class NamedSequenceKey(GetAttrKey, SequenceKey):
    def __str__(self) -> str:
        return f".{self.name}"


class OpTreeTreeLib(AbstractTreeLib):
    @staticmethod
    def map(
        func: Callable[..., Any],
        tree: Any,
        *rest: Any,
        is_leaf: Callable[[Any], bool] | None = None,
        is_parallel: bool | ParallelConfig = False,
    ) -> Any:
        leaves, treedef = ot.tree_flatten(tree, is_leaf, namespace=namespace)
        flat = [leaves] + [treedef.flatten_up_to(r) for r in rest]
        if not is_parallel:
            return ot.tree_unflatten(treedef, [func(*args) for args in zip(*flat)])
        config = dict() if is_parallel is True else is_parallel
        return ot.tree_unflatten(treedef, concurrent_map(func, flat, **config))

    @staticmethod
    def path_map(
        func: Callable[..., Any],
        tree: Any,
        *rest: Any,
        is_leaf: Callable[[Any], bool] | None = None,
        is_parallel: bool | ParallelConfig = False,
    ) -> Any:
        leaves, treedef = ot.tree_flatten(tree, is_leaf, namespace=namespace)
        flat = [leaves] + [treedef.flatten_up_to(r) for r in rest]
        flat = (ot.treespec_paths(treedef), *flat)
        if not is_parallel:
            return ot.tree_unflatten(treedef, [func(*args) for args in zip(*flat)])
        config = dict() if is_parallel is True else is_parallel
        return ot.tree_unflatten(treedef, concurrent_map(func, flat, **config))

    @staticmethod
    def flatten(
        tree: Any,
        *,
        is_leaf: Callable[[Any], bool] | None = None,
    ) -> tuple[Iterable[Any], ot.PyTreeDef]:
        leaves, treedef = ot.tree_flatten(tree, is_leaf, namespace=namespace)
        return (leaves, treedef)

    @staticmethod
    def path_flatten(
        tree: Any,
        *,
        is_leaf: Callable[[Any], bool] | None = None,
    ) -> tuple[Iterable[KeyPathLeaf], ot.PyTreeDef]:
        leaves, treedef = ot.tree_flatten(tree, is_leaf, namespace=namespace)
        return (list(zip(ot.treespec_paths(treedef), leaves)), treedef)

    @staticmethod
    def unflatten(treedef: ot.PyTreeDef, leaves: Iterable[Any]) -> Any:
        return ot.tree_unflatten(treedef, leaves)

    @staticmethod
    def register_treeclass(klass: type[Tree]) -> None:
        def unflatten(keys: tuple[str, ...], leaves: tuple[Any, ...]) -> Tree:
            vars(tree := getattr(object, "__new__")(klass)).update(zip(keys, leaves))
            return tree

        def flatten(tree: Tree):
            dynamic = dict(vars(tree))
            keys = tuple(dynamic.keys())
            entries = tuple(NamedSequenceKey(*ik) for ik in enumerate(keys))
            return (tuple(dynamic.values()), keys, entries)

        ot.register_pytree_node(klass, flatten, unflatten, namespace)

    @staticmethod
    def register_static(klass: type[Tree]) -> None:
        ot.register_pytree_node(klass, lambda x: ((), x), lambda x, _: x, namespace)

    @staticmethod
    def attribute_key(name: str) -> GetAttrKey:
        return GetAttrKey(name)

    @staticmethod
    def sequence_key(index: int) -> SequenceKey:
        return SequenceKey(index)

    @staticmethod
    def dict_key(key: Hashable) -> DictKey:
        return DictKey(key)

    @staticmethod
    def keystr(keys: Any) -> str:
        return "".join(str(key) for key in keys)
